//
// Textflow.java - Console interface
//
// Something that can be called from multiple threads to
//   send packets over the text channel
//   send packets which require responses
// and can be called on a single thread to respond to packets arriving over
// the text channel.
//

//...simports:0:
import java.net.*;
import java.io.*;
import java.util.*;
//...e

public class Textflow
  {
  protected static final String handshake = "Textflow begin";
  protected InputStream is;
  protected PrintStream ps;
  protected boolean child;
  // Helpers
//...skeyOf:2:
protected static String keyOf(int port, int connId)
  {
  return Integer.toString(port)+"/"+Integer.toString(connId);
  }
//...e
//...sencode:2:
protected static String encode(byte[] b, int len)
  {
  String base64 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
  String e = "";
  int i = 0;
  while ( i + 3 <= len )
    {
    int v0 = b[i++] & 0xff;
    int v1 = b[i++] & 0xff;
    int v2 = b[i++] & 0xff;
    e += String.valueOf( base64.charAt(  (v0>>2)          & 0x3f ) );
    e += String.valueOf( base64.charAt( ((v0<<4)|(v1>>4)) & 0x3f ) );
    e += String.valueOf( base64.charAt( ((v1<<2)|(v2>>6)) & 0x3f ) );
    e += String.valueOf( base64.charAt(   v2              & 0x3f ) );
    }
  switch ( len - i )
    {
    case 1:
      {
      int v0 = b[i] & 0xff;
      e += String.valueOf( base64.charAt(  (v0>>2)          & 0x3f ) );
      e += String.valueOf( base64.charAt(  (v0<<4)          & 0x3f ) );
      e += "==";
      }
      break;
    case 2:
      {
      int v0 = b[i++] & 0xff;
      int v1 = b[i  ] & 0xff; 
      e += String.valueOf( base64.charAt(  (v0>>2)          & 0x3f ) );
      e += String.valueOf( base64.charAt( ((v0<<4)|(v1>>4)) & 0x3f ) );
      e += String.valueOf( base64.charAt(  (v1<<2)          & 0x3f ) );
      e += "=";
      }
      break;
    }
  return e;
  }
//...e
//...sdecode:2:
protected static byte[] decode(String s)
  {
  String base64 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
  int i = 0, j = 0;
  byte b[] = new byte[(s.length()/4)*3];
    // Every 4 characters input can produce 3 (or less) output
  while ( i + 4 <= s.length() )
    {
    int v0 = base64.indexOf(s.charAt(i++));
    int v1 = base64.indexOf(s.charAt(i++));
    int v2 = base64.indexOf(s.charAt(i++));
    int v3 = base64.indexOf(s.charAt(i++));
    if ( v0 == -1 || v1 == -1 || v2 == -1 || v3 == -1 )
      return null; // base64 encoded data contains bad character
    if ( v0 == 64 || v1 == 64 )
      break;
    b[j++] = (byte) (  (v0<<2)|(v1>>4)         );
    if ( v2 == 64 )
      break;
    b[j++] = (byte) ( ((v1<<4)|(v2>>2)) & 0xff );
    if ( v3 == 64 )
      break;
    b[j++] = (byte) ( ((v2<<6)| v3    ) & 0xff );
    }
  if ( i < s.length() )
    return null; // Encoded data must be a multiple of 4 characters
  if ( j < b.length )
    {
    byte b2[] = new byte[j];
    System.arraycopy(b, 0, b2, 0, j);
    b = b2;
    }
  return b;
  }
//...e
  // Called by Listeners
//...ssend:2:
public void send(String op, int port, int connId, byte[] b, int len)
  {
  String sender = child ? "c" : "p";
  synchronized ( this )
    {
    if ( b != null )
      ps.println(sender+" "+op+" "+port+" "+connId+" "+encode(b,len));
    else
      ps.println(sender+" "+op+" "+port+" "+connId);
    ps.flush();
    }
  }
//...e
  // Called from Tunnel.main
//...sTextflow:2:
Textflow(InputStream is, OutputStream os, boolean child)
  {
  this.is    = is;
  this.ps    = new PrintStream(os);
  this.child = child;
  String sender = child ? "c" : "p";
  ps.println(sender+" "+handshake);
  ps.flush();
  }
//...e
//...sprocess:2:
public void process(Dictionary<Integer,Listener> listeners, Proxy prx)
  {
  String sender = child ? "c" : "p";
  String other  = child ? "p" : "c";
  InputStreamReader isr = new InputStreamReader(is);
  BufferedReader br = new BufferedReader(isr);
  try
    {
    boolean handshook = false;
    String line;
    while ( (line = br.readLine()) != null )
      if ( handshook )
        {
        StringTokenizer st = new StringTokenizer(line);
        if ( st.nextToken().equals(sender) )
          ; // This is something we sent, that the intermediary program
            // (typically telnet) echoed back to use, so ignore it.
        else
          {
          String op = st.nextToken();
          if ( op.equals("!") )
            System.err.println("Remote: "+line.substring(4));
          else
            {
            int port = Integer.parseInt(st.nextToken());
            int connId = Integer.parseInt(st.nextToken());
            byte[] b;
            if ( st.hasMoreTokens() )
              b = decode(st.nextToken());
            else
              b = null;
            if ( Character.isLowerCase(op.charAt(0)) )
              // From Listener, to Proxy (and then on to ProxyConnection)
              {
              if ( ! prx.handle(op, port, connId, b) )
                error("Noise: "+line);
              }
            else
              // From ProxyConnection to Listener (and then on to Connection)
              {
              Listener l = listeners.get(new Integer(port));
              if ( l != null )
                {
                if ( ! l.handle(op, connId, b) )
                  error("Noise: "+line);
                }
              else
                error("No listener to handle: op="+op+" port="+port+" connId="+connId);
              }
            }
          }
        }
      else
        {
        if ( line.equals(other+" "+handshake) )
          handshook = true;
        else if ( line.equals(sender+" "+handshake) )
          ; // The handshake we sent came back to us, don't bother to display
        else if ( child )
          ; // Don't try to log error, there is no textflow to log it over
        else
          error("Pre-handshake: "+line);
        }
    }
  catch ( IOException e )
    {
    error("Cannot process input stream");
    }
  }
//...e
  // Called from various places
//...serror:2:
public void error(String s)
  {
  if ( child )
    // Send special packet to parent, to get it seen
    synchronized ( this )
      {
      ps.println("c ! "+s);
      ps.flush();
      }
  else
    // Can echo the error
    System.err.println("Local: "+s);
  }
//...e
//...slog:2:
protected static void log(String fn, String s)
  {
  try
    {
    FileOutputStream fos = new FileOutputStream(fn, true);
    PrintStream ps = new PrintStream(fos);
    ps.println(s);
    ps.flush();
    fos.close();
    }
  catch ( FileNotFoundException e )
    {
    }
  catch ( IOException e )
    {
    }
  }
//...e
  }
